﻿'Cellular Automata Class Library
'Copyright (C) 2010, Jörg Meier
'see folder [license] for details.

Option Strict On
Option Infer On

Imports CellularAutomata

Namespace Specialized

    Partial Class Rectangular2DCluster(Of TCell As Rectangular2DCell(Of TCell, TConnection), TConnection As Rectangular2DConnection(Of TCell, TConnection))

        ''' <summary>
        ''' Returns the subset of neighbours having a value not higher than the average value
        ''' </summary>
        ''' <param name="Cell"></param>
        ''' <param name="Neighbours"></param>
        ''' <param name="GetValue"></param>
        ''' <param name="StableDelta"></param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Protected Shared Function FindActiveNeighbourSubset( _
            ByVal Cell As TCell, _
            ByVal Neighbours() As NeighbourInfo, _
            ByVal GetValue As Func(Of TCell, Double), _
            ByVal StableDelta As Double) As NeighbourInfo()

            Dim result() As NeighbourInfo = CType(Neighbours.Clone, NeighbourInfo())
            Dim ResultCount As Integer = result.Length
            Do
                'find average
                Dim Average As Double = GetAverage(Cell:=Cell, neighbours:=result, GetValue:=GetValue)

                'zu hohe Nachbarn raus!
                Dim SomethingDone As Boolean = False
                For j As Integer = result.Length - 1 To 0 Step -1
                    If result(j) Is Nothing Then Continue For
                    Dim NeighbourValue As Double = GetValue.Invoke(result(j).OtherCell)
                    If NeighbourValue > Average Then
                        result(j) = Nothing
                        ResultCount -= 1
                        SomethingDone = True
                    End If
                Next j
                If Not SomethingDone Then Exit Do
            Loop

            'remove cells, if their value is to height
            Dim CellValue As Double = GetValue(Cell)
            For j As Integer = result.Length - 1 To 0 Step -1
                If result(j) Is Nothing Then Continue For
                Dim NeighbourValue As Double = GetValue(result(j).OtherCell)
                If NeighbourValue + StableDelta > CellValue Then
                    result(j) = Nothing
                    ResultCount -= 1
                End If
            Next j

            'create the resulting array
            If ResultCount > 0 Then
                Dim aResult(0 To ResultCount - 1) As NeighbourInfo
                Dim i As Integer = 0
                For j As Integer = result.Length - 1 To 0 Step -1
                    If result(j) Is Nothing Then Continue For
                    aResult(i) = result(j)
                    i += 1
                Next j
                Return aResult
            Else
                Dim aResult(-1) As NeighbourInfo
                Return aResult
            End If
        End Function


        ''' <summary>
        ''' Returns the average value of all given cells
        ''' </summary>
        ''' <param name="Cells"></param>
        ''' <param name="GetValue"></param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Protected Shared Function GetAverage(ByVal Cells As IEnumerable(Of TCell), ByVal GetValue As Func(Of TCell, Double)) As Double
            Dim result As Double = 0
            Dim Count As Integer = 0
            For Each cell As TCell In Cells
                If cell Is Nothing Then Continue For
                Count += 1
                result += GetValue(cell)
            Next
            Return result / Count
        End Function

        ''' <summary>
        ''' Returns the average value of all given cells
        ''' </summary>
        ''' <param name="neighbours"></param>
        ''' <param name="GetValue"></param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Protected Shared Function GetAverage(ByVal neighbours As IEnumerable(Of NeighbourInfo), ByVal GetValue As Func(Of TCell, Double)) As Double
            Dim result As Double = 0
            Dim Count As Integer = 0
            For Each NeighbourInfo As NeighbourInfo In neighbours
                If NeighbourInfo Is Nothing Then Continue For
                Count += 1
                result += GetValue(NeighbourInfo.OtherCell)
            Next NeighbourInfo
            Return result / Count
        End Function

        ''' <summary>
        ''' Returns the average value of the cell and the given neighbours
        ''' </summary>
        ''' <param name="Cell"></param>
        ''' <param name="Neighbours"></param>
        ''' <param name="GetValue"></param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Protected Shared Function GetAverage(ByVal Cell As TCell, ByVal neighbours As IEnumerable(Of NeighbourInfo), ByVal GetValue As Func(Of TCell, Double)) As Double
            Dim result As Double = GetValue(Cell)
            Dim Count As Integer = 1
            For Each NeighbourInfo As NeighbourInfo In neighbours
                If NeighbourInfo Is Nothing Then Continue For
                Count += 1
                result += GetValue(NeighbourInfo.OtherCell)
            Next NeighbourInfo
            Return result / Count
        End Function

        ''' <summary>
        ''' Returns the weighted average value of all given cells
        ''' </summary>
        ''' <param name="Cells"></param>
        ''' <param name="GetValue"></param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Protected Shared Function GetWeightedAverage(ByVal Cells As IEnumerable(Of TCell), ByVal GetValue As Func(Of TCell, Double), ByVal GetWeight As Func(Of TCell, Double)) As Double
            If GetWeight Is Nothing Then Return GetAverage(Cells:=Cells, GetValue:=GetValue)

            Dim result As Double = 0
            Dim WeightSum As Double = 0
            For Each cell As TCell In Cells
                If cell Is Nothing Then Continue For
                Dim CellValue As Double = GetValue(cell)
                Dim CellWeight As Double = GetWeight(cell)
                WeightSum += CellWeight
                result += CellValue * CellWeight
            Next
            Return result / WeightSum
        End Function


        ''' <summary>
        ''' Returns the weighted average value of all given cells
        ''' </summary>
        ''' <param name="neighbours"></param>
        ''' <param name="GetValue"></param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Protected Shared Function GetWeightedAverage(ByVal neighbours As IEnumerable(Of NeighbourInfo), ByVal GetValue As Func(Of TCell, Double), ByVal GetWeight As Func(Of TConnection, Double)) As Double
            If GetWeight Is Nothing Then Return GetAverage(neighbours:=neighbours, GetValue:=GetValue)

            Dim result As Double = 0
            Dim WeightSum As Double = 0
            For Each neighbour As NeighbourInfo In neighbours
                If neighbour Is Nothing Then Continue For
                Dim CellValue As Double = GetValue(neighbour.OtherCell)
                Dim CellWeight As Double = GetWeight(neighbour.Connection)
                WeightSum += CellWeight
                result += CellValue * CellWeight
            Next
            Return result / WeightSum
        End Function


        Protected Shared Function FindActiveNeighbourhood( _
            ByVal Cell As TCell, _
            ByVal Neighbours() As NeighbourInfo, _
            ByVal GetValue As Func(Of TCell, Double), _
            ByVal ConnectionWeight As Func(Of TConnection, Double), _
            ByVal StableDelta As Double, _
            ByVal MaxActiveValue As Double) As NeighbourhoodInfo

            'zu hohe Nachbarzellen ausschließen!
            Dim ActiveNeighbours() As NeighbourInfo = FindActiveNeighbourSubset( _
                Cell:=Cell, _
                Neighbours:=Neighbours, _
                GetValue:=GetValue, _
                StableDelta:=StableDelta)

            If ActiveNeighbours.Length < 1 Then Return Nothing

            Dim AverageValue As Double = GetAverage(Cell:=Cell, neighbours:=ActiveNeighbours, GetValue:=GetValue)
            Dim CellValue As Double = GetValue.Invoke(Cell)

            Dim ActiveValue As Double = CellValue - AverageValue
            If ActiveValue > MaxActiveValue Then ActiveValue = MaxActiveValue
            If ActiveValue <= 0 Then Return Nothing 'diese Zelle hat nichts zu geben!

            'welches Gefälle gibt es zu den Nachbarzellen?
            Dim ValueDelta(0 To ActiveNeighbours.Length - 1) As Double
            Dim ValueDeltaSum As Double = 0
            For j As Integer = 0 To ActiveNeighbours.Length - 1
                Dim Neighbour As NeighbourInfo = ActiveNeighbours(j)
                Dim Delta As Double = CellValue - GetValue.Invoke(Neighbour.OtherCell)
                Delta *= ConnectionWeight(Neighbour.Connection)
                ValueDelta(j) = Delta
                ValueDeltaSum += Delta
            Next j
            If ValueDeltaSum <= 0 Then Return Nothing 'kein Gefälle?!?

            Dim result(0 To ActiveNeighbours.Length - 1) As NeighbourInfoExt
            For j As Integer = 0 To ActiveNeighbours.Length - 1
                Dim Neighbour As NeighbourInfo = ActiveNeighbours(j)
                Dim Delta As Double = ValueDelta(j) '/ ValueDeltaSum
                result(j) = New NeighbourInfoExt(NeighbourInfo:=Neighbour, Delta:=Delta)
            Next

            Return New NeighbourhoodInfo(ActiveNeighbours:=result, ActiveValue:=ActiveValue, Average:=AverageValue)
        End Function






        ''' <summary>
        ''' Returns the Laplace-Operator (Δ) for the given state variable.
        ''' </summary>
        ''' <param name="Cell"> Centeral cell the Laplace-Operator should be computed for </param>
        ''' <param name="GetValue"> Function providing the state variable value for a given cell </param>
        ''' <param name="GetWeight"> Function providing the weighting for given cell. If GetWeight is nothing, 1 is used for all cells. </param>
        ''' <returns></returns>
        ''' <remarks>
        ''' Taken from Rudy Rucker: "Continuous-Valued Cellular Automata in Two Dimensions" in New Constructions in Cellular Automata, edited by David Griffeath and Cristopher Moore, Oxford University Press page 300
        ''' </remarks>
        Protected Shared Function GetLaplaceOperatorScalar( _
            ByVal Cell As TCell, _
            ByVal GetValue As Func(Of TCell, Double), _
            ByVal GetWeight As Func(Of TConnection, Double)) As Double

            Dim Neighbours As IEnumerable(Of NeighbourInfo) = FindNeighbours(Cell)
            Dim NeighbourAverage As Double = GetWeightedAverage(neighbours:=Neighbours, GetValue:=GetValue, GetWeight:=GetWeight)
            Dim CellValue As Double = GetValue.Invoke(Cell)

            Return NeighbourAverage - CellValue
        End Function

        ''' <summary>
        ''' Returns the Laplace-Operator (Δ) for the given state variable and neighbours.
        ''' </summary>
        ''' <param name="Cell"> Centeral cell the Laplace-Operator should be computed for </param>
        ''' <param name="Neighbours"> Enumeration of all neighbours to be includet. </param>
        ''' <param name="GetValue"> Function providing the state variable value for a given cell </param>
        ''' <param name="GetWeight"> Function providing the weighting for given cell. If GetWeight is nothing, 1 is used for all cells. </param>
        ''' <returns></returns>
        ''' <remarks>
        ''' Taken from Rudy Rucker: "Continuous-Valued Cellular Automata in Two Dimensions" in New Constructions in Cellular Automata, edited by David Griffeath and Cristopher Moore, Oxford University Press page 300
        ''' </remarks>
        Protected Shared Function GetLaplaceOperatorScalar( _
            ByVal Cell As TCell, _
            ByVal Neighbours As IEnumerable(Of NeighbourInfo), _
            ByVal GetValue As Func(Of TCell, Double), _
            ByVal GetWeight As Func(Of TConnection, Double)) As Double

            Dim NeighbourAverage As Double = GetWeightedAverage(neighbours:=Neighbours, GetValue:=GetValue, GetWeight:=GetWeight)
            Dim CellValue As Double = GetValue.Invoke(Cell)

            Return NeighbourAverage - CellValue
        End Function

    End Class

End Namespace
