'Cellular Automata Class Library
'Copyright (C) 2010, Jrg Meier
'see folder [license] for details.

Option Strict On
Option Infer On

Imports CellularAutomata

Namespace Specialized

    Partial Class SpecializedCluster(Of TCell As Cell(Of TCell, TConnection), TConnection As Connection(Of TCell, TConnection))

        ''' <summary>
        ''' Represents the method that will create a hexagonal cell.
        ''' </summary>
        ''' <param name="cluster"> Parent cell cluster the cell will be added to. </param>
        ''' <param name="center"> Center coordinates of the cell to create. </param>
        ''' <param name="polygon"> Outline polygon of the cell to create </param>
        ''' <returns> The new hexagonal cell. Retuns Null (Nothing in VisualBasic), if no cell should be added at this location. </returns>
        ''' <remarks></remarks>
        <System.ComponentModel.EditorBrowsable(ComponentModel.EditorBrowsableState.Advanced)>
        Public Delegate Function HexagonalCellFactory(cluster As Cluster(Of TCell, TConnection), center As System.Windows.Point, polygon() As System.Windows.Point) As TCell



        Friend MustInherit Class HexagonalClusterFactory
            Private Sub New()
            End Sub

            ''' <summary>
            ''' Creates a 2D hexagonal array of cells and connections and adds it to the given cluster
            ''' </summary>
            ''' <param name="cluster"> Cluster to add the cells and connections to. </param>
            ''' <param name="CellCountX"> Count of new cells in x-direction. </param>
            ''' <param name="CellCountY"> Count of new cells in y-direction. </param>
            ''' <param name="CellFactory"> Method for creation of cells. </param>
            ''' <param name="ConnectionFactory"> Method for creation of connections. </param>
            ''' <returns> 2D-array of created cells. </returns>
            ''' <remarks></remarks>
            Friend Shared Function AddHexagonalCluster( _
                cluster As SpecializedCluster(Of TCell, TConnection), _
                CellCountX As Integer, CellCountY As Integer, _
                InnerCellRadius As Double,
                CellFactory As HexagonalCellFactory, _
                ConnectionFactory As ConnectionFactory) As TCell(,)

                'http://en.wikipedia.org/wiki/Hexagonal_tiling

                Dim OuterCellRadius = InnerCellRadius * 2 / System.Math.Sqrt(3)
                Dim dx = InnerCellRadius * 2
                Dim dy = InnerCellRadius * System.Math.Sqrt(3)

                'create all cells
                Dim Cells(0 To CellCountX - 1, 0 To CellCountY - 1) As TCell
                For i As Integer = Cells.GetLowerBound(0) To Cells.GetUpperBound(0)
                    For j As Integer = Cells.GetLowerBound(1) To Cells.GetUpperBound(1)

                        Dim x As Double = i * dx
                        Dim y As Double = j * dy
                        If (j Or 1) = j Then x += InnerCellRadius

                        Dim polygon(0 To 5) As System.Windows.Point
                        polygon(0) = New System.Windows.Point(x + InnerCellRadius, y - OuterCellRadius / 2)
                        polygon(1) = New System.Windows.Point(x, y - OuterCellRadius)
                        polygon(2) = New System.Windows.Point(x - InnerCellRadius, y - OuterCellRadius / 2)
                        polygon(3) = New System.Windows.Point(x - InnerCellRadius, y + OuterCellRadius / 2)
                        polygon(4) = New System.Windows.Point(x, y + OuterCellRadius)
                        polygon(5) = New System.Windows.Point(x + InnerCellRadius, y + OuterCellRadius / 2)

                        Dim cell As TCell = CellFactory.Invoke(cluster:=cluster, center:=New System.Windows.Point(x, y), polygon:=polygon)

                        Cells(i, j) = cell
                        If cell IsNot Nothing Then
                            Call cluster.Cells.Add(cell)
                        End If
                    Next
                Next

                'create a array of connection-lists
                Dim CellConnections(0 To Cells.Length - 1) As List(Of TConnection)
                For i As Integer = 0 To CellConnections.Length - 1
                    CellConnections(i) = New List(Of TConnection)
                Next

                'create all needed connections
                For i As Integer = Cells.GetLowerBound(0) To Cells.GetUpperBound(0)
                    For j As Integer = Cells.GetLowerBound(1) To Cells.GetUpperBound(1)
                        Dim Connections As New List(Of TConnection)

                        AddConnection(Connections, Cells, i, j, +1, 0, ConnectionTypes.Primary, ConnectionFactory)
                        AddConnection(Connections, Cells, i, j, 0, +1, ConnectionTypes.Primary, ConnectionFactory)
                        AddConnection(Connections, Cells, i, j, -1, +1, ConnectionTypes.Primary, ConnectionFactory)

                        For Each Connection As TConnection In Connections
                            If Connection Is Nothing Then Continue For
                            If Connection.Cell1 Is Nothing OrElse Connection.Cell2 Is Nothing Then Continue For
                            Call cluster.Connections.Add(Connection)
                            Call CellConnections(Connection.Cell1.Index).Add(Connection)
                            Call CellConnections(Connection.Cell2.Index).Add(Connection)
                        Next Connection
                    Next
                Next

                For i As Integer = 0 To cluster.Cells.Count - 1
                    Call cluster.RegisterCellConnections(cluster.Cells(i), CellConnections(i).ToArray)
                Next

                Return Cells
            End Function

            Private Shared Sub AddConnection(connections As ICollection(Of TConnection),
                                 cells(,) As TCell,
                                 i As Integer, j As Integer,
                                 di As Integer, dj As Integer,
                                 type As ConnectionTypes,
                                 ConnectionFactory As ConnectionFactory)

                If i < cells.GetLowerBound(0) OrElse i > cells.GetUpperBound(0) Then Return
                If j < cells.GetLowerBound(1) OrElse j > cells.GetUpperBound(1) Then Return
                If (i + di) < cells.GetLowerBound(0) OrElse (i + di) > cells.GetUpperBound(0) Then Return
                If (j + dj) < cells.GetLowerBound(1) OrElse (j + dj) > cells.GetUpperBound(1) Then Return

                Dim c1 = cells(i, j)
                Dim c2 = cells(i + di, j + dj)
                If c1 Is Nothing Or c2 Is Nothing Then Return

                Dim connection As TConnection = ConnectionFactory(c1, c2, type)
                If connection IsNot Nothing Then
                    Call connections.Add(connection)
                End If
            End Sub

        End Class

    End Class

End Namespace
